﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;

namespace Unsafeinterleaving
{
    class Program
    {
        private static int account = 0;
        private static object anyobj = new object();
        /// <summary>
        /// add some money to the shared account
        /// and do some local work
        /// </summary>
        /// <param name="incoming"></param>
        /// <param name="howMany"></param>
        private static void Accountant(int incoming, int howMany)
        {
            for (int i = 0; i < howMany; i++)
            {
                account += incoming; 
                Thread.SpinWait(50000);//do nothing while n operations, simulate local work
            }
        }

        /// <summary>
        /// add some money to the shared account
        /// and do some local work, but account operation is in lock
        /// </summary>
        /// <param name="incoming"></param>
        /// <param name="howMany"></param>
        private static void GoodAccountant(int incoming, int howMany)
        {
            for (int i = 0; i < howMany; i++)
            {
                lock (anyobj)
                {
                    account += incoming;
                }
                Thread.SpinWait(50000);//do nothing while n operations, simulate local work
                                       //remember releas lock as soon as it is possible. local operatation should be outer part of lock.
                
            }
        }
        static void Main(string[] args)
        {
            Stopwatch sw = new Stopwatch(); //for time mesuring
            sw.Start();
            ///let acccount it sequential, there is no task. Just account one operation after another
            Console.WriteLine("Lets count...");
            for (int i = 0; i < 8; i++)
            {
                Accountant(1, 1000);
            }
            Console.WriteLine($"We have {account} in {sw.ElapsedMilliseconds} ms"); //it is correct but slow
            
            Console.WriteLine("Ok let some parralelizm...");
            object anyObject = new object();
            account = 0;
            sw.Restart();
            Task[] tasks = new Task[8];
            for (int i = 0; i < 8; i++) //start 8 tasks. All run and calculating on the same variable
            {
                tasks[i] = Task.Factory.StartNew(() =>
                {
                    Accountant(1, 1000); 
                });
            }
            Task.WaitAll(tasks);
            Console.WriteLine($"OMG! We have {account} in {sw.ElapsedMilliseconds} ms"); //it is fast but not always correct

            Console.WriteLine("Ok let some parralelizm with lock...");
            account = 0;
            sw.Restart();
            for (int i = 0; i < 8; i++)
            {
                tasks[i] = Task.Factory.StartNew(() =>
                {
                    lock (anyobj)
                    {
                        Accountant(1, 1000);
                    }
                });
            }
            Task.WaitAll(tasks);
            Console.WriteLine($"Now, we have again {account} again :) in {sw.ElapsedMilliseconds} ms");

            Console.WriteLine("Ok let some parralelizm with proper lock...");
            account = 0;
            sw.Restart();
            for (int i = 0; i < 8; i++)
            {
                tasks[i] = Task.Factory.StartNew(() =>
                {
                        GoodAccountant(1, 1000);
                });
            }
            Task.WaitAll(tasks);
            Console.WriteLine($"Now, we have again {account} again :) in {sw.ElapsedMilliseconds} ms");
                   
        }
    }
}
